*- Name.......Paint-Box Handled Seeker (SeekerPX)
*- Author.....Todd Kreuter
*- For........dBASE Plus, dB2K 

*- Copyright 2002, Todd B Kreuter. All rights reserved.
// This code may be used freely in any dBASE product and 
// distrubuted freely in any dBASE deployed application.
// Any other use and distribution is restricted.

*- Description
// A paintbox handled seeker type control capable of handling
// multiple edit (seeker) controls within the control.

// Form Metric must be Pixels!

// This class is intended to be used as a base class for other
// contols. See SeekerPX.htm for more information.

*- Acknowledgements
// Beta Testers - Seah KC, Chris Stoeckert, and Marc Van den Berghen.
//						Thank you for your time and effort! [Tested the
//						grid combobox from which this class evolved.]

*- History
// 2002.0725 -	Initial
// 2002.1014 - Removed all hard coded edit controls to simplify
	//				the use of this class as a baseClass.
// 2002.1018 - Added LocateNext method, locate property, and
	// 			changed onSearchKey to convert search string to
   //				uppercase when using an indexed search.
// 2002.1212 - Changes for new bookmark property, fix problem
	//				where onChange not firing when edit value blank.
// 2003.0101 - Added refName property for SPXEdit because name was
	//          changing in deployed app to Seeeker or Master.
// 2003.0102 - Changed backspace key behaviour, only reselects
	//				last char, no longer performs a search.

#define EM_SETSEL		0x00B1

class SeekerPX(parentObj) of CONTAINER(parentObj) custom

   *- API for caret in edit control
   if TYPE("CreateCaret") # "FP"
      extern 	CLOGICAL CreateCaret(CHANDLE, CHANDLE, CINT, CINT) User32
   endif

   if TYPE("ShowCaret") # "FP"
      extern 	CLOGICAL ShowCaret(CHANDLE) User32
   endif

   if TYPE("HideCaret") # "FP"
   	extern 	CLOGICAL HideCaret(CHANDLE) User32
   endif
 
   if TYPE("SetCaretPos") # "FP"
   	extern 	CLOGICAL SetCaretPos(CINT, CINT) User32
   endif

	*- API to repaint control, needed when covered by another window
   if TYPE("UpdateWindow") # "FP"
      extern 	CLOGICAL UpdateWindow(CHANDLE) User32
   endif

   *- API for sending various messages
   if TYPE("SendMessageN") # "FP"
      extern   CLONG SendMessageN(CHANDLE, CUINT, CUINT, CLONG) User32 ;
               from "SendMessageA"
   endif

	*- Check for Pixel Metric
	if parentObj.metric # 6
   	MSGBOX("Form Metrics Must Be Pixels!", "Notice", 48)
   endif

   with (this)
   	height	= 22
   	width		= 106
      metric	= 6	// Pixels
   endwith

	*- Internal properties
   this.focus	  	= false  // Whether control has focus
   this.modify		= false	// Whether to modify master edit
   this.locate		= false	// Whether control is in locateMode

   this.rowset		= null	// Reference to dataSource rowset
	this.seeker	  	= null	// Reference to seeker edit control
   this.master		= null	// Reference to master edit control
   this.bookMark	= null	// DataSource rowset bookmark

   this.picture  	= ""		// Format picture of seeker
 	this.cursor	  	= 0		// Cursors current position in seeker
   this.hinge 	  	= 0		// Cursors anchor position in seeker
   this.inserts	= null	// Array of literal char inserts

	*- External properties
	this.pageNext 	= 1      // Rows navigated by PageUp/Down
   this.findKey	= false	// Whether to use indexed search findKey()
   this.rangeKey	= ""		// Key constraining rowset with setRange()

	*- Event methods
   this.canSelChange	= {||True}
	this.onSelChange	= {||Null}
   this.onEditSel	 	= {||Null}

	function onDesignOpen
   *********************
   *- Force anchored controls to reposition

   	this.width ++
      this.width --

      return

	function beforeOpen
   *******************
	*- Instantiate control before opening

      local objFirst, objNow, objSeek
      private cMacro

      *- Setup SPXEdit Controls
      objFirst = this.first
      objNow   = this.first
      objSeek	= null

      if objNow # null
         do
            if objNow.ClassName = "SPXEDIT" AND NOT EMPTY(objNow.dataLink)

					objNow.refName = (objNow.name)

            	if objNow.enabled AND objNow.visible
            		objNow.onLeftMouseDown	= class::onMouseDown
               endif

               if objNow.dataSource = null
               	objNow.dataSource = objNow.dataLink
               endif

					if objSeek = null AND objNow.visible AND objNow.enabled
               	objSeek = objNow
               endif

               if this.master = null
               	this.master = objNow
               endif

            endif

            objNow = ObjNow.before

         until ObjNow = ObjFirst
      endif

		*- DataSource rowset derived from master edit
      if TYPE("this.master.dataSource") == "O"
	      this.rowset	= this.master.dataSource.parent.parent
      endif

      *- Check If Can Enable Control
      if objSeek = null
      	MSGBOX("No Seeker Control Found!", "SeekerPX", 48)
         this.enabled = false

      elseif this.rowset = null
      	MSGBOX("No DataSource Rowset Defined!", "SeekerPX", 48)
         this.enabled = false
      endif

		if this.enabled

         *- keyHandler Paintbox
         this.keyHandler = new PaintBox(this)
         with (this.keyHandler)
            onGotFocus	= class::KH_onGotFocus
            onLostFocus	= class::KH_onLostFocus
            onKeyDown	= class::KH_onKeyDown
            onKeyUp		= class::KH_onKeyUp
            onChar		= class::KH_onChar
            width    	= 0
            height   	= 0
         endwith

         this.keyHandler.copy			= class::copy
         this.keyHandler.shiftDown	= false
         this.keyHandler.ctrlDown 	= false

         *- textSel Overlay
         this.textSel = new TextLabel(this)
         with (this.textSel)
            onLeftMouseDown	= class::onMouseDown
            height				= 14
            left 					= 0
            top 					= 0
            width 				= 0
            metric 				= 6	// Pixels
            text 					= ""
            colorNormal 		= "HighLightText/HighLight"
            fontSize				= 8
            mousePointer 		= 3	// IBeam
            prefixEnable 		= false
         endwith

         this.textSel.start 	= 0
         this.textSel.end	 	= 0

         *- Timer
         this.timer = new timer()
         this.timer.parent    	= this
         this.timer.onTimer   	= this.onTimer
         this.timer.interval		= 0.005
         this.timer.enabled		= false
         this.timer.message		= REPLICATE( CHR(0), 14) // Msg Struc

         *- Assign canSelChange function pointer
         if type(this.form.className + "::" + this.name + "_canSelChange") = "FP"
            cMacro = this.form.className + "::" + this.name + "_canSelChange"
            this.canSelChange = &cMacro.
         endif

         *- Assign onSelChange function pointer
         if type(this.form.className + "::" + this.name + "_onSelChange") = "FP"
            cMacro = this.form.className + "::" + this.name + "_onSelChange"
            this.onSelChange = &cMacro.
         endif

         *- Assign onEditSel function pointer
         if type(this.form.className + "::" + this.name + "_onEditSel") = "FP"
            cMacro = this.form.className + "::" + this.name + "_onEditSel"
            this.onEditSel = &cMacro.
         endif

         this.setSeeker(objSeek)

		endif

      return

	function setFocus(bFocus)
   *****************
   	if PCOUNT() == 0
      	bFocus = true
      endif

      if NOT this.focus AND bFocus			// Gaining Focus
      	this.focus		= true

			UpdateWindow(this.hwnd)
         this.keyHandler.setFocus()
         this.timer.enabled = true
			this.rowset.locateOptions = 3

         this.onGotFocus()

      elseif this.focus AND NOT bFocus		// Losing Focus
			this.timer.enabled = false
         this.focus = false

	      HideCaret(this.seeker.hwnd)
         this.textSel.text		= ""
	      this.textSel.width	= 0

         this.onLostFocus()
      endif

      if this.focus								// Activate Seeker
	      CreateCaret(this.seeker.hwnd, null, 1, 14)
   	   ShowCaret(this.seeker.hwnd)
         this.seeker.valueSel = this.seeker.value
	      this.setCursorSel(0, -1)
      endif

		this.locate = false

   	return

	function onGotFocus
   *******************
   *- Provided for subclass.

   	return

   function onLostFocus
   ********************
   * Provided for subclass.

   	return

	function KH_onGotFocus
   **********************
   *- Handler Method: Call parent setFocus.

   	if NOT this.parent.focus
	      this.parent.setFocus(true)
      endif

      return

   function KH_onLostFocus
   ***********************
   *- Handler Method: Call parent setFocus.

     	return this.parent.setFocus(false)

   function KH_onChar(nKey, nRepCnt, nFlags)
   ******************
   *- Handler Method: Printable Key Cases

      if (nKey > 31 AND nKey < 128)
      	this.parent.onSearchKey(nKey)

      endif

      return

   function KH_onKeyDown(nKey, nRepCnt, nFlags)
   *********************
	*- Handler Method: Keydown Cases

      do case
 		case nKey = 16								// Shift
      	this.shiftDown = true

      case nKey = 17						 		// Ctrl
         this.ctrlDown = true

		case nKey = 8
	      this.parent.SetCursorSel(MAX(this.parent.textSel.start - 1, 0), -1, true)
      	
      case nKey = 38 OR nKey = 40
      	this.parent.onNavigateKey(nKey)

      case (nKey > 34 AND nKey < 40) AND NOT this.ctrlDown
      	this.parent.onCursorKey(nKey)

      case nKey = 37 AND this.ctrlDown		// ArrowLeft
        	this.parent.nextEdit(-1)
         this.parent.setFocus(true)

      case nKey = 39 AND this.ctrlDown	 	// ArrowRight
        	this.parent.nextEdit(1)
         this.parent.setFocus(true)

		endcase

		return false

   function KH_onKeyUp(nKey, nRepCnt, nFlags)
   *******************
   *- Handler method: KeyUp cases, including PageUp/Down
   // because they are not recognized in KeyDown (BUG).

      do case
      case nKey = 16								// Shift
         this.shiftDown	= false

      case nKey = 17								// Ctrl
         this.ctrlDown = false

      case nKey = 33 OR nKey = 34
      	this.parent.onNavigateKey(nKey, this.CtrlDown)

      case nKey = 67 AND this.ctrlDown		// Ctrl + C (Copy)
      	this.parent.copy()

      endcase

      return false

	function onMouseDown(nFlag, nCol, nRow)
   ********************
   *- Edit and textSel method. Sets focus to control, and/or
   // sets the clicked edit as current, or calls onCursorKey
   // to position cursor based on clicked column.

   	if this.name = "TEXTSEL"
      	nCol += (this.left - this.parent.seeker.left)

         this.parent.onCursorKey(0, nCol)

      elseif this.className = "SPXEDIT"
      	if this.hwnd # this.parent.seeker.hwnd
         	local objEdit; objEdit = this
         	this.parent.setSeeker(objEdit)
		     	this.parent.setFocus(true)

         elseif NOT this.parent.focus
         	this.parent.setFocus(true)

         else
         	this.parent.onCursorKey(0, nCol)
         endif

      endif

	function onTimer
   ****************
   *- Refresh current edit for changes made to dataLink not
   // initiated by control (rowset navigation and abandon)

      if this.parent.seeker.value # this.parent.seeker.valueSel
         this.parent.seeker.valueSel = this.parent.seeker.value
         this.parent.setCursorSel(0, -1)
      endif

   	return

	function onSearchKey(nKey)
   ********************
	*- Performs an incremental search in dataSource rowset on
   //  edit input string, called from handler onChar.

   	if NOT this.canChange(true)
      	return false
      endif

      local nStart, nLength, bFind, i, cInput, cField
      nStart 	= this.textSel.start
      nLength	= this.textSel.end - nStart
      bFind		= false

      *- Adjust nStart for literal character inserts
      for i = 1 to this.inserts.size
         if nStart >= this.inserts[i]
            nStart --
         endif
      endfor

      cInput	= this.seeker.value.rightTrim().stuff(nStart, nLength, CHR(nKey))
      nStart	= this.textSel.start + 1

      HideCaret(this.seeker.hwnd)

      this.rowset.notifyControls = false

      if this.findKey
         bFind = this.rowset.findKey( UPPER(this.rangeKey + cInput) )
         this.locate = false

      else
         cField = this.seeker.dataSource.fieldName

         try
            bFind = this.rowset.applyLocate(cField + " = [" + cInput + "]")
         catch(exception e)
            // do nothing
         endtry

         this.locate = bFind

      endif

      this.rowset.notifyControls = true

      if bFind
         if this.bookMark # this.rowset.bookmark()
            this.rowset.next(0)
            this.onChange()
         endif

//	      this.textSel.visible = false			//TEST
         this.setCursorSel(nStart, -1, nKey = 8)
//	      this.textSel.visible = true			//TEST

      elseif this.bookmark # null
         this.rowset.goto(this.bookmark)
      endif

      ShowCaret(this.seeker.hwnd)

      return

	function LocateNext
   *******************
   *- Calls the dataSource rowset locateNext method. Only
   // applicable when using non-indexed searches and control
   // is in locateMode (locate property is true).
   // Called from handler onKeyDown. 

   	if NOT this.canChange(true)
      	return false
      endif

   	local bFind
      bFind	= false

		if this.locate

      	this.rowset.notifyControls = false

         try
            bFind = this.rowset.locateNext()
         catch(exception e)
            // do nothing
         endtry

	      this.rowset.notifyControls = true

         if bFind
            if this.bookMark # this.rowset.bookmark()
               this.rowset.next(0)
               this.onChange()
            endif

            this.setCursorSel(this.textSel.start, -1)

         elseif this.bookMark # null
            this.rowset.goto(this.bookMark)
         endif

         this.locate = bFind

	      ShowCaret(this.seeker.hwnd)

      endif

      return

	function onNavigateKey(nKey, bCtrl)
   **********************
   *- Navigates the dataSource rowset based on key code,
   // called from handler onKeyDown and onKeyUp.
    
   	if NOT this.canChange(nKey # 0)
      	return false
      endif

   	local bNext
      bNext	= true

      HideCaret(this.seeker.hwnd)

      do case
      case nKey	= 38							// ArrowUp
      	this.rowset.next(-1)
         bNext = false

      case nKey	= 40				  			// ArrowDn
      	this.rowset.next(1)

      case nKey	= 33 AND bCtrl				// PageUp + Ctrl (Home)
      	this.rowset.first()

      case nKey	= 33							// PageUp
         this.rowset.next(-this.pageNext)
         bNext = false

      case nKey	= 34 AND bCtrl				// PageDn + Ctrl (End)
			this.rowset.last()

      case nKey	= 34							// PageDn
      	this.rowset.next(this.pageNext)

      case nKey	= 0							// Special
      	this.rowset.next(0)
         this.bookmark = null

      endcase

		if this.rowset.endofset
      	if bNext
         	this.rowset.last()
         else
         	this.rowset.first()
         endif
      endif

      if this.bookmark # this.rowset.bookMark()
      	this.onChange()
      endif

      this.setCursorSel(0, -1)
      ShowCaret(this.seeker.hwnd)

      return

	function onCursorKey(nKey, nCol)
   ********************
   *- Sets a new cursor position and selects text if shiftDown,
   // otherwise hinge and textSel are reset to cursor. Called
   // from handler onKeyDown and mouse click event.

   *- Params:	nKey		- Key Handler Key Code, 0 for Mouse
   //				nCol		- Mouse column clicked

		local cValue, nCursor, bLeft
      cValue	= TRANSFORM(this.seeker.value, this.picture).rightTrim()
      nCursor	= 0
      bLeft		= false

      do case
      case nKey	= 0								// Mouse
      	nCursor = this.CharFromCol(nCol)

      case nKey	= 35								// End
         nCursor	= -1

      case nKey	= 36 								// Home
         nCursor	= 0
         bLeft		= true

      case nKey	= 37								// ArrowLeft
      	nCursor	= MAX(this.cursor - 1, 0)
         bLeft		= true

      case nKey	= 39								// ArrowRight
      	nCursor	= this.cursor + 1

      endcase

      this.cursor = this.validCursorPos(nCursor, cValue, bLeft)

      if this.keyHandler.shiftDown
         this.setTextSel(cValue)
      else
         this.hinge				= this.cursor
         this.textSel.text		= ""
         this.textSel.start	= this.cursor
         this.textSel.end		= this.cursor
			this.textSel.width 	= 0
	      SetCaretPos(this.ColFromChar(this.cursor, cValue), 0)
      endif

      return

	function canChange(bSync)
   ******************
   *- Check if selection can be changed and ensure dataSource
   // rowset is sycnchronized with master edit value.

   *- Params: bSync	- Whether to sychronize rowset.

   	private bChange
     	bChange = this.canSelChange()

      if TYPE("bChange") # "L"
         bChange = false
      endif

      // Sychronize rowset
      if bChange

			if EMPTY(this.master.value)
         	this.bookmark = null

         elseif bSync AND this.master.value # this.master.dataSource.value

            local cField, cValue, bFind
            cField	= this.master.dataSource.fieldName
            cValue	= this.master.value.rightTrim()

				try
					if this.findKey
               	bFind = this.rowset.findKey(cValue)

               else
               	bFind = this.rowset.applyLocate(cField + " = [" + cValue + "]")
               endif

            catch(exception e)
               // do nothing
            endtry

	      	this.bookmark = IIF(bFind, this.rowset.bookmark(), null)

         else
         	this.bookmark = this.rowset.bookmark()

         endif

      endif

      return bChange

   function onChange
   *****************
	*- Update the master edit value from its dataSource
   // when the dataSource rowset selection has changed.

      local cValue
      cValue = this.master.dataSource.value

   	if this.modify
         this.master.value = cValue
		endif

      this.onSelChange(cValue)

      this.seeker.valueSel = this.seeker.value

      return

	function setCursorSel(nStart, nEnd, bLeft)
   *********************
   *- Sets hinge and cursor to valid cursor positions, calls
   // setTextSel to perform text selection in edit control.

   *- Params:	nStart	- Start of selection range (hinge)
   //				nEnd		- End of selection range	(cursor)
   //				bLeft		- Validate nStart to left	(moving left)

   	local cValue
		cValue = TRANSFORM(this.seeker.value, this.picture).rightTrim()

      this.hinge	= this.validCursorPos(nStart, cValue, bLeft)
      this.cursor	= this.validCursorPos(nEnd, cValue, false)

      this.setTextSel(cValue)

      return

	function validCursorPos(nCursor, cString, bLeft)
   ***********************
   *- Returns a valid cursor position in string, excludes
   // literal character inserts from picture format.

   *- Params:	nCursor	- Cursor positon to validate
   //				cString	- Cursors string
   //				bLeft		- Validate cursor to left (moving left)

   	local nValid
      nValid = nCursor

      if nCursor = -1
      	nValid = cString.length
      endif

      do while this.inserts.scan(nValid) > 0
      	if bLeft
         	nValid --

         else
         	nValid ++
         endif
      enddo

      nValid = MIN(MAX(nValid, 0), cString.length)

      return nValid

	function setTextSel(cString)
   *******************
   *- Sets the textSel properties from the hinge and cursor,
   // and selects range of text within edit value (cString).
   // Uses the textSel textLabel to overlay the result.

   	local nLeft, nRight, bLeft
      nLeft 	= MIN(this.cursor, this.hinge)
      nRight	= MAX(this.cursor, this.hinge)
      bLeft		= (nLeft = this.cursor)

      this.textSel.text 	= cString.subString(nLeft, nRight)
      this.textSel.start	= nLeft
      this.textSel.end		= nRight

		nLeft		= this.ColFromChar(nLeft, cString)
      nRight	= MIN(this.ColFromChar(nRight, cString), this.seeker.width)

      this.textSel.left		= nLeft + this.seeker.left
      this.textSel.width	= MAX(nRight - nLeft, 0)

		if bLeft
	      SetCaretPos(nLeft - 1, 0)

      else
      	SetCaretPos(nRight, 0)
      endif

		return

	function colFromChar(nChar, cString)
   ********************
   *- Returns the x coordinate of the specified character index
   // in a string using getTextExtent.

   	local nColumn
      nColumn	= 2

      if nChar # 0
	   	nColumn = INT(this.textSel.getTextExtent(cString.left(nChar)) / 0.1429) + 3
		endif

      return nColumn


	function charFromCol(nCol)
   ********************
   *- Returns the character index of the character nearest the
   // specified x coordinate in a string using getTextExtent.

   	local cString, nChar
      cString = TRANSFORM(this.seeker.value, this.picture).rightTrim()

      for nChar = 0 to cString.length
      	if INT(this.textSel.getTextExtent(cString.left(nChar)) / 0.1429) + 3 > nCol
         	exit
         endif
      endfor

		return MAX(nChar - 1, 0)

	function nextEdit(nNext)
   *****************
   *- Selects next edit control relative to the current edit,
   // or seeker, and sets it as the new seeker.

		local aEdits, objFirst, objNow, nEdit, nSize
      aEdits	= new array()
      objFirst = this.first
      objNow   = this.first
      nEdit		= 0
      nSize		= 0

      if objNow # null
         do
            if objNow.ClassName = "SPXEDIT" AND objNow.visible AND objNow.enabled
	          	aEdits.add(objNow)
               nSize ++

               if objNow = this.seeker
               	nEdit = nSize + nNext
               endif
            endif

            objNow = objNow.before

         until objNow = objFirst
      endif

     	if nEdit > aEdits.size
	    	nEdit = 1

      elseif nEdit < 1
      	nEdit = aEdits.size
      endif

      if aEdits[nEdit] # this.seeker
         this.setSeeker(aEdits[nEdit])
      endif

      return

	function setSeeker(objEdit)
   ******************
   *- Sets objEdit as the current edit (seeker).
	// Fires the onEditSel method.

      // Cleanup current seeker
      if TYPE("this.seeker.hwnd") == "N"
         HideCaret(this.seeker.hwnd)
         this.textSel.text		= 0
         this.textSel.width	= 0
      endif

		this.seeker	  	= objEdit
      this.picture	= ""
      this.inserts  	= new array()

      if NOT EMPTY(this.seeker.function)
         this.picture = "@" + this.seeker.function + " "
      endif

      if NOT Empty(this.seeker.picture)
         this.picture += this.seeker.picture

         for i = 0 to this.seeker.picture.length - 1
            if NOT this.seeker.picture.substring(i, i + 1) $ "9|#|N|A|X"
               this.inserts.add(i)
            endif
         endfor
      endif

      this.textSel.top			= this.seeker.top
      this.textSel.fontName	= this.seeker.fontName
      this.textSel.fontSize	= this.seeker.fontSize

		this.onEditSel(this.seeker.refName)

		return

	function copy
   *************
   	local nStart, nEnd
      nStart = this.textSel.start
      nEnd	 = this.textSel.end

      if nStart # nEnd
			SendMessageN(this.seeker.hwnd, EM_SETSEL, nStart, nEnd)
         this.seeker.copy()
      endif

      return

	function setIndex(cIndexName, bFindKey)
   *****************
   *- Sets the rowset indexName and findKey flag.
   *- The bFindKey param is optional and defaults to true
   // when an indexName is passed, and false if otherwise.

   	try
			if PCOUNT() < 1
         	cIndexName = ""
         endif

      	if PCOUNT() < 2
         	bFindKey = NOT EMPTY(cIndexName)
         endif

      	this.rowset.indexName = cIndexName
         this.findKey = bFindKey

      catch(exception e)
			MSGBOX(e.message, "SetIndex", 48)
         this.findKey = false
      endtry

      return

endclass

class SPXEdit(parentObj) of ENTRYFIELD(parentObj)
*- SeekerPX Edit control.
// Protecting border until SeekerPX is modified to handle the 
// various borders, as the positioning of the textSel control
// would be affected by the border.

	protect border, borderStyle
    
   with (this)
		when			= {||False}
      height		= 16
      left			= 0
      top			= 0
      width			= 0
      metric		= 6	// Pixels
      fontSize 	= 8
      value			= "Edit"
      borderStyle = 3	// None
   endwith

   this.dataSource	= null	// dataSource field object
   this.valueSel		= ""		// Value selected via control
   this.refName		= ""		// Original reference name

endclass

