/*

	Filename		 :	Volumecontrol.cc
   Author		 :	Marc Van den Berghen
   Date			 :	08-10-2002

   Update		 : 12-10-2002 Ronnie MacGregor
                  Class re-written to give self contained drop on form
                  functionality. Now derived from slider instead of paintbox,
                  with defaults to suit Jean-Pierre Martel's CD Player.


   Purpose		 : To control the volume of the soundcard from within
                  a dbase-application

	Dependencies : The file volume.dll must be in scope. (either you put it in
                  the same directory as the cc, or you have to load it from
                  another directory load dll "path...\volume.dll"

	Usage 	    : The volumeconrol is designed as a slider-control, which can be
                  dropped from the custom tab of the component palette.It will 
                  look automatically for the maximum and minimum-values of the
                  soundcard volume and map it to values in the range from 0 to
                  100, 0 being mute and 100 being max-volume. To set the volume
                  use the method setmastervolume(nvol).

                  eg. // set the vol to half of its maximum
                      form.volumecontrol1.setmastervolume(50) 

                  The dll contains a callback-function, that is used to notify
                  the volume control of changes, by generating a rightmousedown  
                  event. You can design your own special-looking control, set
                  the slider invisible, and just refer to its value and methods.

   Note      	 : The winmm.dll-(windows multimedia extension) supports
                  computers with more then one soundcard. However, the
                  volumecontrol.cc will ALWAYS use the first one in install
                  order. No way to control a second or third one at least from
                  within dBL ...

*/


#define MM_MIXM_CONTROL_CHANGE  0x3D1
#define GWL_WNDPROC             (-4)


class VolumeControl(parentobj) of SLIDER(parentobj) custom

   // Ensure control drops onto designer surface properly
   form.metric = 6 // Pixels

   with (this)
   	onrightmousedown = class::update
		canclose = {; mixerclose(this.mastervolume.getulong(20)); this.mastervolume.release()}   
      onChange = {; this.SetMasterVolume(this.value)}
		ongotfocus = {; this.parent.setfocus()}
      speedTip = "Volume control"
      height = 15
      width = 110
      value = 50
      colorNormal = "BtnFace"
      mousePointer = 13	// Hand
      rangeMax = 100
      rangeMin = 0
      tics = 2	// None
      vertical = false
      enableSelection = false
      visible = false
   endwith


   function onOpen

      // initialization of the API-calls
      extern culong 		mixersetup(cptr,chandle) volume.dll
      extern culong 		setvol(cptr) volume.dll
      extern cuint  	  	GetWindowLong(chandle,cint) user32 from "GetWindowLongA"
      extern clong  	  	SetWindowLong(chandle,cint,clong) user32 from "SetWindowLongA"
      extern CHANDLE    GetWindow(chandle,cUINT) user32 
      extern clogical 	SetPropA(chandle, cptr, CHANDLE) user32 from "SetPropA"
      extern CHANDLE 	GetProp(Chandle, cptr) user32 from "GetPropA"
      extern cLONG    	Adroffunc(cuint) volume.dll
      extern cuint 		SendMessage(chandle,cuint,clong,clong) user32 from "SendMessageA"
      extern cuint		mixerClose(chandle) winmm.dll

      // setup the callback-function
      this.oldfunc=getwindowlong(this.hwnd,GWL_WNDPROC) //address of the Window-Procedure
      setpropa(this.hwnd,"O",this.oldfunc)
      setpropa(this.hwnd,"A",this.hwnd)
      newfunc=adroffunc(1)
      setwindowlong(this.hwnd,GWL_WNDPROC,newfunc)

      // Instantiate Structure as a custom property         
      this.mastervolume = new mixerstruct()

      // setup the audiomixer
      mixersetup(this.mastervolume,this.hwnd)

      // Set the initial value of the slider
      sendmessage(this.hwnd,MM_MIXM_CONTROL_CHANGE,0,0) 

      // Set the relative volume factor
      this.mastervolume.InitialiseRelativeVolumeFactor()          

      // Display the initialised control
      this.visible=true 

      return                                   


function SetMasterVolume(n)
   n=int(n)
   // check if value in range
   n=iif(n<0,0,iif(n>100,100,n)) 
   this.mastervolume.setrel(n)
   setvol(this.mastervolume)
return

function update(nflags,col,row)
   // this function is called each time the volume is changed
   local n
   n=int(nflags/this.mastervolume.getlong(16))
   this.value=n
return

endclass


//////////////////


class mixerstruct of string

   /*
      A structure, to hold the following items:

      minvalue - at index 0
      minimum value of the soundcardvolume (mute)

      maxvalue - at index 4
      highest possible volume

      volume - at index 8
      The actual value of the mastervolume in "soundcard-metric"

      control volume - at index 12
      The actual value of this control (0-100)

      relvol - at index 16
      relative volume conversion factor

      mixerhandle - at index 20
      like a window handle but for audiocontrols

      The first two are used to calculate the relvol conversion factor 

   */    

	this.string=space(12)

   function getLong(nIndex)
      local n
      n = this.getByte(nIndex) + ;
         bitlshift(this.getByte(nIndex + 1), 8) + ;
         bitlshift(this.getByte(nIndex + 2), 16) + ;
         bitlshift(this.getByte(nIndex + 3), 24)
      if bitset(n, 31)
         n := - bitnot(n)
      endif
      return n

   function getuLong(nIndex)
      local n
      n = this.getByte(nIndex) + ;
         bitlshift(this.getByte(nIndex + 1), 8) + ;
         bitlshift(this.getByte(nIndex + 2), 16) + ;
         bitlshift(this.getByte(nIndex + 3), 24)
      return n

   function setLong(nIndex, nValue)
      local bRet
      try
         bRet = true
         this.setByte(nIndex, bitand(nValue, 0xFF))
         this.setByte(nIndex + 1, bitrshift(bitand(nValue, 0xFF00), 8))
         this.setByte(nIndex + 2, bitrshift(bitand(nValue, 0xFF0000), 16))
         this.setByte(nIndex + 3, bitrshift(bitand(nValue, 0xFF000000), 24))
      catch(exception e)
         ? e.message, e.lineno
         bRet = false
      endtry
      return bRet

   function release
      this.string = ''
      release object this
      return
      
   function setrel(nvalue)
   	this.setlong(12,nvalue)
      this.setlong(8,nvalue*this.getlong(16))
      return

   function InitialiseRelativeVolumeFactor()                          
      this.setlong(16,int((this.getlong(4)-this.getlong(0))/100))     
      return                                                          

endclass

